use std::{
    fs::File,
    io::{Read, Seek, SeekFrom},
    os::mikros::{fs::FileExt, ipc, syscalls, Errno, FileOpenMode},
    path::Path,
    sync::{atomic::{AtomicU64, Ordering}, Arc},
};

use dashmap::DashMap;
use itertools::Itertools;
use sharded_slab::Slab;
use uuid::Uuid;

#[derive(Clone, Copy, Debug, Default)]
#[allow(unused)]
struct MbrEntry {
    active: bool,
    typ: u8,
    start: u32,
    size: u32,
}

#[derive(Clone, Debug, Default)]
#[allow(unused)]
struct GptEntry {
    type_guid: Uuid,
    entry_guid: Uuid,
    start: u64,
    len: u64,
    name: String,
}

struct OpenFile {
    drive_file: (u64, u64),
    offset: u64,
    part_len: usize,
    pos: AtomicU64,
}

impl Clone for OpenFile {
    fn clone(&self) -> Self {
        Self {
            drive_file: self.drive_file,
            offset: self.offset,
            part_len: self.part_len,
            pos: AtomicU64::new(self.pos.load(Ordering::Relaxed)),
        }
    }
}

#[derive(Clone)]
struct Serv {
    parts: Arc<DashMap<String, ((u64, u64), u64, usize)>>,
    open_files: Arc<Slab<OpenFile>>,
}

impl dev_driver_rpc::Server for Serv {
    fn open(&self, path: &std::path::Path, _mode: FileOpenMode) -> Result<u64, Errno> {
        let part = *(self.parts.get(path.to_str().unwrap()).ok_or(Errno::EBADF)?);
        Ok(self.open_files.insert(OpenFile {
            drive_file: part.0,
            offset: part.1,
            part_len: part.2,
            pos: AtomicU64::new(0),
        }).unwrap() as u64)
    }
}

impl file_rpc::Server for Serv {
    fn read(&self, fd: u64, len: usize) -> Result<std::borrow::Cow<'_, [u8]>, Errno> {
        let file = &*self.open_files.get(fd as usize).ok_or(Errno::EBADF)?;
        if file.pos.load(Ordering::Relaxed).saturating_add(len as u64) > file.part_len as u64 {
            return Err(Errno::EINVAL);
        }

        let file_client = file_rpc::Client::new(file.drive_file.0);
        file_client.seek(file.drive_file.1, file_rpc::SeekFrom::Start(file.offset + file.pos.load(Ordering::Relaxed)))?;
        Ok(file_client
            .read(file.drive_file.1, len)?
            .into())
    }

    fn write(&self, fd: u64, data: &[u8]) -> Result<(), Errno> {
        let file = &*self.open_files.get(fd as usize).ok_or(Errno::EBADF)?;
        if file.pos.load(Ordering::Relaxed).saturating_add(data.len() as u64) > file.part_len as u64 {
            return Err(Errno::EINVAL);
        }

        let file_client = file_rpc::Client::new(file.drive_file.0);
        file_client.seek(file.drive_file.1, file_rpc::SeekFrom::Start(file.offset + file.pos.load(Ordering::Relaxed)))?;
        Ok(file_client
            .write(file.drive_file.1, &data)?
            .into())
    }

    fn close(&self, _fd: u64) -> Result<(), Errno> { Ok(()) }

    fn size(&self, fd: u64) -> Result<u64, Errno> {
        Ok(self.open_files.get(fd as usize).ok_or(Errno::EBADF)?.part_len as u64)
    }

    fn dup(&self, fd: u64) -> Result<u64, Errno> {
        Ok(fd)
    }

    fn seek(&self, fd: u64, pos: file_rpc::SeekFrom) -> Result<u64, Errno> {
        let file = &*self.open_files.get(fd as usize).unwrap();
        match pos {
            file_rpc::SeekFrom::Start(offset) => {
                file.pos.store(offset, Ordering::Relaxed);
            },
            file_rpc::SeekFrom::End(offset) => {
                if offset <= 0 {
                   file.pos.store(file.part_len as u64 - (-offset) as u64, Ordering::Relaxed);
                }
            },
            file_rpc::SeekFrom::Current(offset) => {
                if offset > 0 {
                    file.pos.store(u64::min(file.pos.load(Ordering::Relaxed) + offset as u64, file.part_len as u64), Ordering::Relaxed);
                } else {
                    let offset = (-offset) as u64;
                    if offset > file.pos.load(Ordering::Relaxed) {
                        file.pos.store(0, Ordering::Relaxed);
                    } else {
                        file.pos.fetch_sub(offset, Ordering::Relaxed);
                    }
                }
            },
        }
        Ok(file.pos.load(Ordering::Relaxed))
    }
}

fn parse_mbr(mut disk: &File) -> Vec<MbrEntry> {
    let mut mbr = [0; 512];

    disk.read_exact(&mut mbr).unwrap();

    (0..4)
        .map(|i| {
            let entry_offset = 0x1BE + 0x10 * i;
            let active = mbr[entry_offset] & 0x80 == 0x80;
            let typ = mbr[entry_offset + 0x4];
            let start = u32::from_le_bytes(
                mbr[entry_offset + 0x8..entry_offset + 0xC]
                    .try_into()
                    .unwrap(),
            );
            let size = u32::from_le_bytes(
                mbr[entry_offset + 0xC..entry_offset + 0x10]
                    .try_into()
                    .unwrap(),
            );

            MbrEntry {
                active,
                typ,
                start,
                size,
            }
        })
        .collect_vec()
}

fn parse_gpt(mut disk: &File) -> Result<(Uuid, Vec<GptEntry>), ()> {
    let mut gpt_header = [0; 512];

    disk.seek(SeekFrom::Start(512)).unwrap();
    disk.read_exact(&mut gpt_header).unwrap();

    if &gpt_header[0..8] != b"EFI PART" {
        return Err(());
    }

    let disk_guid = Uuid::from_bytes_le(gpt_header[56..72].try_into().unwrap());
    let part_table_start_lba = u64::from_le_bytes(gpt_header[72..80].try_into().unwrap()) as usize;
    let num_parts = u32::from_le_bytes(gpt_header[80..84].try_into().unwrap()) as usize;
    let entry_len = u32::from_le_bytes(gpt_header[84..88].try_into().unwrap()) as usize;

    let part_table_len = num_parts * entry_len;

    let mut gpt_part_table = vec![0; part_table_len.next_multiple_of(512)];

    disk.seek(SeekFrom::Start((part_table_start_lba * 512) as u64))
        .unwrap();
    disk.read_exact(&mut gpt_part_table).unwrap();

    let mut parts = Vec::new();

    for i in 0..num_parts {
        let entry_offset = i * entry_len;
        let entry_data = &gpt_part_table[entry_offset..(entry_offset + entry_len)];
        if entry_data[0..16] == [0; 16] {
            continue;
        }
        let type_guid = Uuid::from_bytes_le(entry_data[0..16].try_into().unwrap());
        let entry_guid = Uuid::from_bytes_le(entry_data[16..32].try_into().unwrap());
        let start = u64::from_le_bytes(entry_data[32..40].try_into().unwrap());
        let end = u64::from_le_bytes(entry_data[40..48].try_into().unwrap());
        let len = end - start;
        let name = &entry_data[56..]
            .iter()
            .tuples::<(_, _)>()
            .map(|(&x, &y)| u16::from_le_bytes([x, y]))
            .collect_vec();
        let name = String::from_utf16(name)
            .unwrap()
            .trim_matches([' ', '\0'])
            .to_owned();

        parts.push(GptEntry {
            type_guid,
            entry_guid,
            start,
            len,
            name,
        })
    }

    Ok((disk_guid, parts))
}

fn main() {
    syslog_msg_ipc::register_callback();

    let syslog_pid = loop {
        if let Some(pid) = syscalls::try_get_registered(2) {
            break pid;
        }
    };
    let devfs_pid = loop {
        if let Some(pid) = syscalls::try_get_registered(1) {
            break pid;
        }
    };

    let syslog_client = syslog_rpc::Client::new(syslog_pid);

    syslog_client.subscribe_to_binary("blockdev".to_string(), vec![0]);

    syslog_client
        .send_text_message("partserv", "Partition table server initialized")
        .unwrap();

    let parts = Arc::new(DashMap::new());

    let serv = Serv {
        parts: Arc::clone(&parts),
        open_files: Arc::new(Slab::new()),
    };

    dev_driver_rpc::register_server(Box::new(serv.clone()));
    file_rpc::register_server(Box::new(serv));

    loop {
        ipc::process_messages();
        while let Some(msg) = syslog_msg_ipc::try_get_syslog_msg() {
            if msg.from == "blockdev" {
                if let Some(binary) = msg.binary {
                    if binary.kind != 0 {
                        continue;
                    }
                    let dev_name = std::str::from_utf8(binary.data.as_slice()).unwrap();

                    let disk = File::open(format!("/dev/{}", dev_name)).unwrap();

                    let mbr_entries = parse_mbr(&disk);

                    if mbr_entries[0].typ == 0xEE {
                        let Ok((disk_guid, gpt_entries)) = parse_gpt(&disk) else {
                            syslog_client
                                .send_text_message(
                                    "partserv",
                                    format!("/dev/{dev_name}: GPT table has invalid magic!"),
                                )
                                .unwrap();
                            continue;
                        };
                        syslog_client
                            .send_text_message(
                                "partserv",
                                format!("/dev/{dev_name}: GPT table, UUID {disk_guid:X}"),
                            )
                            .unwrap();

                        let drive_file = std::fs::File::open(Path::new(&format!("/dev/{}", dev_name))).unwrap().get_pid_fd();

                        for (i, part) in gpt_entries.iter().enumerate() {
                            let part_dev_name = format!("{dev_name}p{i}");

                            syslog_client
                        .send_text_message(
                            "partserv",
                            format!(
                                "/dev/{part_dev_name}: UUID {:X}, start {}, size {}, type {:X}, name {:?}",
                                part.entry_guid,
                                humansize::format_size(part.start * 512, humansize::BINARY),
                                humansize::format_size(part.len * 512, humansize::BINARY),
                                part.type_guid,
                                part.name,
                            ),
                        )
                        .unwrap();

                            parts.insert(
                                part_dev_name.clone(),
                                (drive_file, part.start * 512, (part.len * 512) as usize),
                            );

                            devfs_rpc::Client::new(devfs_pid)
                                .register_dev(&part_dev_name)
                                .unwrap();
                        }
                    } else {
                        println!("Disk has MBR");

                        let drive_file = std::fs::File::open(Path::new(&format!("/dev/{}", dev_name))).unwrap().get_pid_fd();

                        for (i, part) in mbr_entries.iter().enumerate() {
                            if part.size == 0 {
                                continue;
                            }
                            let act_str = if part.active { ", active" } else { "" };

                            syslog_client
                                .send_text_message(
                                    "partserv",
                                    format!(
                                    "/dev/{dev_name}p{i}: start {}, size {}, type {:#x}{act_str}",
                                    humansize::format_size(part.start * 512, humansize::BINARY),
                                    humansize::format_size(part.size * 512, humansize::BINARY),
                                    part.typ,
                                ),
                                )
                                .unwrap();

                            let part_dev_name = format!("{dev_name}p{i}");

                            parts.insert(
                                part_dev_name.clone(),
                                (
                                    drive_file,
                                    (part.start * 512) as u64,
                                    (part.size * 512) as usize,
                                ),
                            );

                            devfs_rpc::Client::new(devfs_pid)
                                .register_dev(&part_dev_name)
                                .unwrap();
                        }
                    }
                }
            }
        }
    }
}
